#
# Rhythmbox scaletempo plugin
#
import rhythmdb, rb
import gtk
import gst
import math

# A utility class that holds a measure of pitch scaling
# and is able to convert between different units.
class PitchScale:
	def __init__(self):
		self.pitch = 1

	def set_fraction(self, val):
		self.pitch = val
	def get_fraction(self):
		return self.pitch

	def set_percent(self, val):
		self.pitch = val/100
	def get_percent(self):
		return self.pitch*100

	def set_semitone(self, val):
		self.pitch = math.pow(1.059463094, val)
	def get_semitone(self):
		return math.log(self.pitch, 1.059463094)

	def set_indexed(self, val, index):
		if index == 0:
			return self.set_percent(val)
		if index == 1:
			return self.set_semitone(val)
		return self.set_fraction(val)
	def get_indexed(self, index):
		if index == 0:
			return self.get_percent()
		if index == 1:
			return self.get_semitone()
		return self.get_fraction()

	# returns (min, default, max, fine_step, course_step, display_fmt)
	def get_adjdata(self, index):
		if index == 0:
			# percent
			return (0, 100, 200, 1, 10, '% 5d');
		if index == 1:
			# semitone
			return (-12, 0, 12, 0.1, 1, '%+.1f');
		return None

# Classes for managing pitch scaling GStreamer plugin this is to allow
# different plugins to be used.
class GstPitchScaler:
	def __init__(self):
		pass

	def activate(self, player):
		player.add_filter(self.gstelement)
		self.player = player

	def deactivate(self):
		self.player.remove_filter(self.gstelement)
		del self.player
		del self.gstelement

	def set_tempo_fraction(self, val):
		pass
	def get_tempo_fraction(self):
		return None

	def set_pitch_fraction(self, val):
		pass
	def get_pitch_fraction(self):
		return None

class GstPitchScalerPitch (GstPitchScaler):
	def __init__(self):
		GstPitchScaler.__init__(self)
		registry = gst.registry_get_default()
		if gst.element_factory_find('pitch') == None:
			raise ImportError, 'GStreamer element "pitch" not available.'

	def activate(self, player):
		self.gstelement = gst.element_factory_make('pitch', 'pitch0')
		GstPitchScaler.activate(self, player)

	def set_tempo_fraction(self, val):
		val = max(0.1, val)
		val = min(10, val)
		self.gstelement.set_property('tempo', val)
		return self.get_tempo_fraction()
	def get_tempo_fraction(self):
		return self.gstelement.get_property('tempo')

	def set_pitch_fraction(self, val):	
		val = max(0.1, val)
		val = min(10, val)
		self.gstelement.set_property('pitch', val)
		return self.get_pitch_fraction()
	def get_pitch_fraction(self):
		return self.gstelement.get_property('pitch')


# The Rhythmbox plugin
class ScaletempoPlugin (rb.Plugin):
	def __init__(self):
		rb.Plugin.__init__(self)

	def activate(self, shell):
		print "activating scaletempo plugin"
		self.shell = shell
		# find available pitch scaling GStreamer element
		try:
			self.scaler = GstPitchScalerPitch()
		except ImportError, e:
			# try other methods here when available
			# last one raises an error
			# TODO better error message
			rb.error_dialog(title="Couldn't load scaletempo plugin",
			                message="No suitable GStreamer element found.")
			return False

		# Rhythmbox can hang if pipeline is changed while playing, so stop first
		shell.get_player().stop()
		self.scaler.activate(shell.get_player().get_property('player'))
		# add control to toolbar
		builder = gtk.Builder()
		builder.add_from_file(self.find_file('scaletempo.xml'))
		self.uicomponent = builder.get_object('container')
		self.scale_tempo = builder.get_object('scale-tempo')
		self.label_tempo = builder.get_object('label-tempo')
		self.scale_pitch = builder.get_object('scale-pitch')
		self.label_pitch = builder.get_object('label-pitch')
		self.combo_pitch = builder.get_object('combo-pitch')
		self.reset = builder.get_object('reset')
		shell.add_widget(self.uicomponent, rb.SHELL_UI_LOCATION_MAIN_TOP)
		# connect signals
		self.scale_tempo.connect('change-value', self.tempo_changed)
		self.scale_pitch.connect('change-value', self.pitch_changed)
		self.combo_pitch.connect('changed', self.combo_pitch_changed)
		self.reset.connect('clicked', self.reset_pressed)

		# add UI overlay (menu items)
		uiman = shell.get_player().get_property('ui-manager')
		self.actgrp = gtk.ActionGroup('ScaletempoPluginActions')
		self.actgrp.add_actions([
			('Scaletempo',        None, _('Tempo and pitch')),

			('tempo-up-coarse',   None, _('Increase tempo by 10%'),
				'<Control>Right', None, lambda a: self.inc_tempo_fraction(0.1)),
			('tempo-up-fine',     None, _('Increase tempo by 1%'),
				'<Control><Shift>Right', None, lambda a: self.inc_tempo_fraction(0.01)),
			('tempo-reset',       None, _('Reset tempo to original'),
				'<Control><Shift>T', None, lambda a: self.set_tempo_fraction(1)),
			('tempo-down-fine',   None, _('Decrease tempo by 1%'),
				'<Control><Shift>Left', None, lambda a: self.inc_tempo_fraction(-0.01)),
			('tempo-down-coarse', None, _('Decrease tempo by 10%'),
				'<Control>Left', None, lambda a: self.inc_tempo_fraction(-0.1)),

			('pitch-up-coarse',   None, _('Increase pitch by a semitone'),
				'<Control>Page_Up', None, lambda a: self.inc_pitch_semitone(1)),
			('pitch-up-fine',     None, _('Increase pitch by 0.1 semitone'),
				'<Control><Shift>Page_Up', None, lambda a: self.inc_pitch_semitone(0.1)),
			('pitch-reset',       None, _('Reset pitch to original'),
				'<Control><Shift>P', None, lambda a: self.set_pitch_fraction(1)),
			('pitch-down-fine',   None, _('Decrease pitch by 0.1 semitone'),
				'<Control><Shift>Page_Down', None, lambda a: self.inc_pitch_semitone(-0.1)),
			('pitch-down-coarse', None, _('Decrease pitch by a semitone'),
				'<Control>Page_Down', None, lambda a: self.inc_pitch_semitone(-1)),
		])

		uiman.insert_action_group(self.actgrp, 0)
		uiman.add_ui_from_file(self.find_file('scaletempo-ui.xml'))

		self.pitch = PitchScale()
		self.combo_pitch.set_active(0)
		self.set_pitch_fraction(1)
		self.set_tempo_fraction(1)
		uiman.ensure_update()


	def deactivate(self, shell):
		# Rhythmbox can hang if pipeline is changed while playing, so stop first
		shell.get_player().stop()

		self.scaler.deactivate()
		self.uicomponent.hide_all()
		del self.actgrp
		del self.reset
		del self.combo_pitch
		del self.scale_pitch
		del self.label_pitch
		del self.scale_tempo
		del self.label_tempo
		del self.uicomponent
		del self.shell
		del self.pitch
		del self.scaler

	###
	### UI Callbacks
	###

	def reset_pressed(self, button):
		self.set_tempo_fraction(1)
		self.set_pitch_fraction(1)

	def combo_pitch_changed(self, foo):
		curval = self.pitch.get_indexed(self.combo_pitch.get_active())
		adjdata = self.pitch.get_adjdata(self.combo_pitch.get_active())
		adj = gtk.Adjustment()
		adj.set_all(curval, adjdata[0], adjdata[2], adjdata[3], adjdata[4], 0)
		self.scale_pitch.set_adjustment(adj)
		# set pitch to update controls
		self.set_pitch_fraction(self.get_pitch_fraction())

	def pitch_changed(self, range, scroll, val):
		self.pitch.set_indexed(val, self.combo_pitch.get_active())
		self.set_pitch_fraction(self.pitch.get_fraction())

	def tempo_changed(self, range, scroll, val):
		self.set_tempo_fraction(val/100)

	###
	### Pitch and tempo manipulation, updates UI and GStreamer plugin
	###

	def set_tempo_fraction(self, val):
		val = self.scaler.set_tempo_fraction(val)
		self.scale_tempo.set_value(val*100)
		self.label_tempo.set_text('%d' % (val*100))
		## gray out menu items if reached tempo min/orig/max
		# note that these vales are different from the scrollbar's to enable
		# a larger range to be set and still have a convenient UI. The values
		# here are from the limits of the GStreamer element.
		# TODO get extrema from GStreamer element class
		self.actgrp.get_action('tempo-up-coarse').set_sensitive(val<=10)
		self.actgrp.get_action('tempo-up-fine').set_sensitive(val<=10)
		self.actgrp.get_action('tempo-reset').set_sensitive(val!=1)
		self.actgrp.get_action('tempo-down-fine').set_sensitive(val>=0.1)
		self.actgrp.get_action('tempo-down-coarse').set_sensitive(val>=0.1)
	def get_tempo_fraction(self):
		return self.scaler.get_tempo_fraction()
	def inc_tempo_fraction(self, val):
		self.set_tempo_fraction(self.get_tempo_fraction()+val)

	def set_pitch_fraction(self, val):
		val = self.scaler.set_pitch_fraction(val)
		adjdata = self.pitch.get_adjdata(self.combo_pitch.get_active())
		self.pitch.set_fraction(val)
		self.scale_pitch.set_value(self.pitch.get_indexed(self.combo_pitch.get_active()))
		self.label_pitch.set_text(adjdata[5] % self.pitch.get_indexed(self.combo_pitch.get_active()))
		# gray out menu items if reached pitch min/orig/max
		self.actgrp.get_action('pitch-up-coarse').set_sensitive(val<=10)
		self.actgrp.get_action('pitch-up-fine').set_sensitive(val<=10)
		self.actgrp.get_action('pitch-reset').set_sensitive(val!=1)
		self.actgrp.get_action('pitch-down-fine').set_sensitive(val>=0.1)
		self.actgrp.get_action('pitch-down-coarse').set_sensitive(val>=0.1)
	def get_pitch_fraction(self):
		return self.scaler.get_pitch_fraction()
	def inc_pitch_fraction(self, val):
		self.set_pitch_fraction(self.get_pitch_fraction()+val)
	def inc_pitch_semitone(self, val):
		self.pitch.set_semitone(self.pitch.get_semitone()+val)
		self.set_pitch_fraction(self.pitch.get_fraction())

